package org.bitbucket.bradleysmithllc.etlunit.integration_test;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.inject.Injector;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.ClassUtils;
import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.feature.AbstractFeature;
import org.bitbucket.bradleysmithllc.etlunit.feature.Feature;
import org.bitbucket.bradleysmithllc.etlunit.feature.RuntimeOption;
import org.bitbucket.bradleysmithllc.etlunit.feature.debug.ConsoleFeatureModule;
import org.bitbucket.bradleysmithllc.etlunit.feature.debug.RunAllFeatureModule;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.parser.*;
import org.bitbucket.bradleysmithllc.etlunit.regexp.PackNameExpression;
import org.bitbucket.bradleysmithllc.etlunit.test_support.MockProcessExecutor;
import org.bitbucket.bradleysmithllc.etlunit.test_support.ScriptMockCallback;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;
import org.bitbucket.bradleysmithllc.etlunit.util.JSonBuilderProxy;
import org.bitbucket.bradleysmithllc.etlunit.util.JsonEscaper;
import org.bitbucket.bradleysmithllc.etlunit.util.VelocityUtil;
import org.junit.Rule;
import org.junit.rules.TemporaryFolder;

import javax.inject.Inject;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.net.URL;
import java.util.*;

public class BaseIntegrationTest
{
	@Rule
	public TemporaryFolder temporaryFolder = new TemporaryFolder();

	protected File root;
	protected File tmp;
	protected File rpt;
	protected File src;
	protected File genSrc;
	protected File cfg;
	protected File rsc;
	protected ETLTestVM etlTestVM;
	private TestResults results;
	protected String identifier;

	private final String thisClassName;

	protected final Map<String, Object> velocityContext = new HashMap<String, Object>();

	protected RuntimeSupport runtimeSupport;

	protected BaseIntegrationTest()
	{
		thisClassName = ClassUtils.getShortClassName(getClass());
	}

	@Inject
	public final void receiveRuntimeSupport(RuntimeSupport support)
	{
		runtimeSupport = support;
	}

	protected TestResults startTest()
	{
		return startTest(getClass().getSimpleName());
	}

	protected TestResults startTest(String testId)
	{
		if (testId == null || testId.trim().equals(""))
		{
			throw new IntegrationTestException("Must provide a valid test id");
		}

		String classKey = thisClassName;

		if (identifier != null)
		{
			classKey += "_" + identifier;
		}

		if (echoLog())
		{
			System.setProperty("org.bitbucket.bradleysmithllc.etlunit.feature.logging.echo", "true");
		}
		else
		{
			System.clearProperty("org.bitbucket.bradleysmithllc.etlunit.feature.logging.echo");
		}

		String resourceRoot = "/integration_tests/" + testId;

		URL url = getClass().getResource(resourceRoot);

		if (url == null)
		{
			throw new IntegrationTestException("Integration test root '" + resourceRoot + "' not present.");
		}

		try
		{
			File intRoot = new File(url.toURI());

			System.out.println("Processing integration test '" + testId + "' in root " + intRoot);

			File confRoot = new File(intRoot, "conf");
			File confJson = new File(confRoot, "etlunit.json");

			if (root == null || cleanWorkspace())
			{
				root = useStaticTestRoot();

				if (root == null)
				{
					File tempFile;

					String prop = System.getProperty("projectBuildDirectory");

					File target = new File("target");

					if (prop != null)
					{
						tempFile = new File(prop, "etlunit-test").getAbsoluteFile();
					}
					else if (target.exists())
					{
						tempFile = new File(target, "etlunit-test").getAbsoluteFile();
					}
					else
					{
						tempFile = temporaryFolder.getRoot();
					}

					root = new File(tempFile, getClass().getName().replace('.', '_') + ".working");
				}

				src = new FileBuilder(root).subdir("src").subdir("test").subdir("etlunit").file();
				cfg = new FileBuilder(root).subdir("src").subdir("test").subdir("resources").subdir("config").file();
				rsc = new FileBuilder(root).subdir("src").subdir("main").file();
				tmp = new FileBuilder(root).subdir("target").subdir("etlunit-temp").mkdirs().file();
				rpt = new FileBuilder(root).subdir("target").file();
				genSrc = new FileBuilder(tmp).subdir("generated-sources").file();

				// create a test source root
				if (root.exists())
				{
					IOUtils.purge(root, true);
				}

				//assertTrue(root.mkdirs());
				assertTrue(src.mkdirs());
				assertTrue(cfg.mkdirs());
				assertTrue(tmp.mkdirs());
				assertTrue(rsc.mkdirs());
				assertTrue(genSrc.mkdirs());
			}

			assertTrue(src.exists());
			assertTrue(cfg.exists());
			assertTrue(rsc.exists());
			assertTrue(tmp.exists());
			assertTrue(genSrc.exists());

			// copy all proprties into the velocity context
			velocityContext.clear();

			velocityContext.put("JsonEscaper", JsonEscaper.class);
			velocityContext.put("junit_temp_dir", temporaryFolder.getRoot());

			velocityContext.put("root_dir", JsonEscaper.escape(root.getAbsolutePath()));
			velocityContext.put("tmp_dir", JsonEscaper.escape(tmp.getAbsolutePath()));
			velocityContext.put("rpt_dir", JsonEscaper.escape(rpt.getAbsolutePath()));
			velocityContext.put("src_dir", JsonEscaper.escape(src.getAbsolutePath()));
			velocityContext.put("gen_src_dir", JsonEscaper.escape(genSrc.getAbsolutePath()));
			velocityContext.put("cfg_dir", JsonEscaper.escape(cfg.getAbsolutePath()));
			velocityContext.put("rsc_dir", JsonEscaper.escape(rsc.getAbsolutePath()));
			velocityContext.put("identifier", JsonEscaper.escape(identifier));
			velocityContext.put("test_class_name", JsonEscaper.escape(thisClassName));

			JSonBuilderProxy bpf =
					new JSonBuilderProxy()
							.object()
							.key("project-root-directory")
							.value(root.getAbsolutePath())
							.key("test-sources-directory")
							.value(src.getAbsolutePath())
							.key("vendor-binary-directory")
							.value(tmp.getAbsolutePath())
							.key("configuration-directory")
							.value(cfg.getAbsolutePath())
							.key("resource-directory")
							.value(rsc.getAbsolutePath())
							.key("reference-directory")
							.value(src.getAbsolutePath())
							.key("temp-directory")
							.value(tmp.getAbsolutePath())
							.key("reports-directory")
							.value(rpt.getAbsolutePath())
							.key("generated-source-directory")
							.value(genSrc.getAbsolutePath())
							.key("runtimeSupport")
							.object()
							.key("processExecutor")
							.value(LoggingProcessExecutor.class.getName())
							.endObject()
							.key("project-name")
							.value("feature-test")
							.key("project-version")
							.value("1.0-test")
							.key("project-user")
							.value("buildUser")
							.key("project-uid")
							.value("buildUid")
							.key("features")
							.object()
							.endObject()
							.endObject();

			ETLTestValueObject etlValue = ETLTestParser.loadObject(bpf.toString());

			if (confJson.exists())
			{
				etlValue = etlValue.merge(ETLTestParser.loadObject(FileUtils.readFileToString(confJson)), ETLTestValueObject.merge_type.right_merge);
			}

			// write out prettilly and run through velocity
			File resultFile = new File(cfg, confJson.getName());

			String jsonContent = new ObjectMapper().writerWithDefaultPrettyPrinter().writeValueAsString(etlValue.getJsonNode());

			String newJsonConfig = VelocityUtil.writeTemplate(jsonContent, velocityContext);

			FileUtils.write(resultFile, newJsonConfig);

			etlTestVM = new ETLTestVM(
					new Configuration(ETLTestParser.loadObject(FileUtils.readFileToString(resultFile)))
			);

			List<RuntimeOption> ro = getRuntimeOptionOverrides();

			if (ro != null)
			{
				etlTestVM.getRuntimeSupport().overrideRuntimeOptions(ro);
			}

			velocityContext.put("runtimeSupport", etlTestVM.getRuntimeSupport());

			etlTestVM.addFeature(new ConsoleFeatureModule());
			etlTestVM.addFeature(new RunAllFeatureModule());
			// add another feature to provide a context callback
			etlTestVM.addFeature(new DebugCallbacker());

			List<Feature> features = getTestFeatures();

			if (features != null)
			{
				Iterator<Feature> it = features.iterator();

				while (it.hasNext())
				{
					etlTestVM.addFeature(it.next());
				}
			}

			etlTestVM.installFeatures();

			URL processUrl = getClass().getResource("/" + classKey);
			if (processUrl != null)
			{
				runtimeSupport.installProcessExecutor(new MockProcessExecutor(new ScriptMockCallback(classKey)));
			}

			File srcRoot = new File(intRoot, "src");

			if (srcRoot.exists())
			{
				// scan the src directory
				deepscanSrc(srcRoot, null, null, scan_type.src);
			}

			File rsrcRoot = new File(intRoot, "feature_src");

			if (rsrcRoot.exists())
			{
				// scan the src directory
				deepscanSrc(rsrcRoot, null, null, scan_type.feature_src);
			}

			inspectTestRootPreTests();
			return etlTestVM.runTests();
		}
		catch (Exception e)
		{
			throw new RuntimeException(e);
		}
	}

	protected void inspectTestRootPreTests()
	{
	}

	enum scan_type
	{
		src,
		feature_src
	}

	;

	private void deepscanSrc(File intRoot, final String package_, final String path, final scan_type stype)
	{
		intRoot.listFiles(new FileFilter()
		{
			@Override
			public boolean accept(File pathname)
			{
				if (pathname.isDirectory())
				{
					PackNameExpression pne = new PackNameExpression(pathname.getName());

					if (pne.matches())
					{
						// if we are in a path, this is illegal
						if (path != null)
						{
							throw new IllegalStateException("Cannot nest a package within a path: " + pathname.getAbsolutePath());
						}

						// this is a package - recurse into
						deepscanSrc(pathname,
								package_ == null ? pne.getPackName() : (package_ + "." + pne.getPackName()),
								path,
								stype);
					}
					else
					{
						// recurse into this without looking for packages
						deepscanSrc(pathname,
								package_,
								path == null ? pathname.getName() : (path + "/" + pathname.getName()),
								stype);
					}
				}
				else
				{
					try
					{
						// process this file to the correct location
						switch (stype)
						{
							case src:
								if (path != null)
								{
									createResource(package_, path, pathname.getName(), FileUtils.readFileToString(pathname));
								}
								else
								{
									createSource(package_, pathname.getName(), FileUtils.readFileToString(pathname));
								}
								break;
							case feature_src:
								// split the first directory off as the feature name
								if (path == null || path.equals(""))
								{
									throw new IntegrationTestException("Cannot load feature source files without a feature directory");
								}

								int index = path.indexOf("/");

								String feature = path;

								if (index != -1)
								{
									feature = path.substring(0, index);
								}

								String subPath = null;

								if (!feature.equals(path))
								{
									subPath = path.substring(index + 1);
								}

								createFeatureSource(feature, subPath, pathname.getName(), FileUtils.readFileToString(pathname));
								break;
						}
					}
					catch (IOException e)
					{
						e.printStackTrace();
					}
				}

				return false;
			}
		});
	}

	protected  List<Feature> getTestFeatures()
	{
		return Collections.EMPTY_LIST;
	}

	protected List<RuntimeOption> getRuntimeOptionOverrides()
	{
		return null;
	}

	private void assertTrue(boolean delete)
	{
		if (!delete)
		{
			throw new IllegalStateException("");
		}
	}

	protected File useStaticTestRoot()
	{
		return null;
	}

	protected boolean echoLog()
	{
		return false;
	}

	protected boolean cleanWorkspace()
	{
		return false;
	}

	private class DebugCallbacker extends AbstractFeature
	{
		int operationNumber = 0;
		int testNumber = 0;

		@Override
		public StatusReporter getStatusReporter()
		{
			return new NullStatusReporterImpl()
			{
				@Override
				public void testCompleted(ETLTestMethod method, CompletionStatus status)
				{
					assertMethodCompletionStatus(method, status, testNumber, operationNumber);
				}
			};
		}

		@Override
		public ClassListener getListener()
		{
			return new NullClassListener()
			{
				@Override
				public void begin(ETLTestClass cl, VariableContext context)
						throws TestAssertionFailure, TestExecutionError, TestWarning
				{
					testNumber = 0;
					assertVariableContextPre(cl, context);
				}

				@Override
				public void begin(ETLTestMethod mt, VariableContext context)
						throws TestAssertionFailure, TestExecutionError, TestWarning
				{
					operationNumber = 0;
					assertVariableContextPre(mt, context, testNumber);
					assertVariableContextPre(mt, context);
					testNumber++;
				}

				@Override
				public void end(ETLTestMethod mt, VariableContext context)
						throws TestAssertionFailure, TestExecutionError, TestWarning
				{
					assertVariableContextPost(mt, context, testNumber, operationNumber);
					assertVariableContextPost(mt, context, operationNumber);
				}

				@Override
				public void end(ETLTestOperation op, ETLTestValueObject parameters, VariableContext vcontext, ExecutionContext econtext)
						throws TestAssertionFailure, TestExecutionError, TestWarning
				{
					assertVariableContextOperation(op, parameters, vcontext, operationNumber);
					operationNumber++;
				}

				@Override
				public void end(ETLTestClass cl, VariableContext context)
						throws TestAssertionFailure, TestExecutionError, TestWarning
				{
					assertVariableContextPost(cl, context, testNumber);
				}
			};
		}

		@Override
		public void initialize(Injector inj)
		{
			inj.injectMembers(BaseIntegrationTest.this);
		}

		@Override
		public String getFeatureName()
		{
			return "debug-caller";
		}

		@Override
		public long getPriorityLevel()
		{
			return Long.MAX_VALUE;
		}
	}

	protected final void createVendorConfig(String vendorName, String name, String content) throws IOException
	{
		File file = new FileBuilder(runtimeSupport.getFeatureConfigurationDirectory(vendorName)).name(name).file();

		try {
			FileUtils.write(file, VelocityUtil.writeTemplate(content, velocityContext));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	protected final void createTemp(String name, String path, String content) throws IOException
	{
		StringBuffer stb = new StringBuffer(content);

		FileBuilder file = new FileBuilder(runtimeSupport.getTempDirectory());

		if (path != null)
		{
			file = file.subdir(path).mkdirs();
		}
		file.name(name);

		try {
			FileUtils.write(file.file(), VelocityUtil.writeTemplate(content, velocityContext));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	protected final void createTemp(String name, String content) throws IOException
	{
		createTemp(name, null, content);
	}

	protected final void createGeneratedSource(String feature, String name, String content) throws IOException
	{
		createGeneratedSource(feature, null, name, content);
	}

	protected final File getGeneratedSource(String feature, String subdir, String name) throws IOException
	{
		FileBuilder fb = new FileBuilder(runtimeSupport.getGeneratedSourceDirectory(feature));
		if (subdir != null)
		{
			fb = fb.subdir(subdir).mkdirs();
		}

		fb = fb.name(name);

		return fb.file();
	}

	protected final void createGeneratedSource(String feature, String subdir, String name, String content)
			throws IOException
	{
		try {
			FileUtils.write(getGeneratedSource(feature, subdir, name), VelocityUtil.writeTemplate(content, velocityContext));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	protected final void createSource(String name, String content) throws IOException
	{
		createSource(null, name, content);
	}

	protected final void createSource(String package_, String name, String content) throws IOException
	{
		try {
			FileUtils.write(getSource(package_, name), VelocityUtil.writeTemplate(content, velocityContext));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	protected final File getSource(String package_, String name) throws IOException
	{
		FileBuilder fb = new FileBuilder(
				package_ != null ? runtimeSupport.getTestSourceDirectory(package_) : runtimeSupport.getTestSourceDirectory()
		);

		File dir = fb.mkdirs().file();

		File file = new File(dir, name);

		return file;
	}

	protected final void createFeatureSource(String feature, String name, String content) throws IOException
	{
		createFeatureSource(feature, null, name, content);
	}

	protected final void createFeatureSource(String feature, String path, String name, String content) throws IOException
	{
		try {
			FileUtils.write(getFeatureSource(feature, path, name), VelocityUtil.writeTemplate(content, velocityContext));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	protected final File getFeatureSource(String feature, String path, String name) throws IOException
	{
		FileBuilder fb = new FileBuilder(runtimeSupport.getFeatureSourceDirectory(feature));

		if (path != null)
		{
			fb = fb.subdir(path);
		}

		File dir = fb.mkdirs().file();

		File file = new File(dir, name);

		return file;
	}

	protected final void createResource(String _package, String relativePath, String name, String content)
			throws IOException
	{
		File
				featSource =
				new FileBuilder(runtimeSupport.getTestResourceDirectory(_package, relativePath)).mkdirs().file();

		File file = new File(featSource, name);

		try {
			FileUtils.write(file, VelocityUtil.writeTemplate(content, velocityContext));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	protected void assertMethodCompletionStatus(ETLTestMethod method, StatusReporter.CompletionStatus status, int testNumber, int operationNumber)
	{
	}

	protected void assertVariableContextOperation(ETLTestOperation op, ETLTestValueObject obj, VariableContext context, int operationNumber)
	{
	}

	protected void assertVariableContextPre(ETLTestMethod mt, VariableContext context)
	{
	}

	protected void assertVariableContextPre(ETLTestMethod mt, VariableContext context, int methodNumber)
	{
	}

	protected void assertVariableContextPre(ETLTestClass cl, VariableContext context)
	{
	}

	protected void assertVariableContextPost(ETLTestMethod mt, VariableContext context, int methodNumber, int operationsProcessed)
	{
	}

	protected void assertVariableContextPost(ETLTestMethod mt, VariableContext context, int operationsProcessed)
	{
	}

	protected void assertVariableContextPost(ETLTestClass cl, VariableContext context, int testsProcessed)
	{
	}
}